/*
 * Copyright (c) 2012-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Thu Aug  2 2012
 */
#include <sys/types.h>
#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <pthread.h>
#include <unistd.h>

#include <cerrno>

#include "ros/init.h"
#include "ros/master.h"
#include "ros/callback_queue.h"
#include "ros/callback_queue_interface.h"

#include "genom3/c/client.h"
#include "genom-iface.h"
#include "private.h"


/* --- local data ---------------------------------------------------------- */

#define ros_cbevent	'e'
#define ros_end		'q'

static int ros_nodeinit = 0;
static int ros_evpipe[2] = { -1, -1 };
static pthread_t ros_evtid;
static int ros_evtid_init = 0;

static void *	ros_evtask(void *data);

static inline genom_event
make_genom_syserr(int code, genom_context self) {
  genom_syserr_detail d = { code };
  return genom_syserr(&d, self);
}


/* --- ros_init ------------------------------------------------------------ */

genom_event
ros_init(genom_client h, int argc, char *argv[])
{
  int s;

  /* global ros init */
  if (!ros_nodeinit) {
    try {
      ros::init(argc, argv, "genom_c_client",
                ros::init_options::NoSigintHandler |
                ros::init_options::AnonymousName |
                ros::init_options::NoRosout);
      if (!ros::master::check()) {
        genom_mwerr_detail d;
        snprintf(d.what, sizeof(d.what), "roscore may not be running on %s:%u",
                 ros::master::getHost().c_str(), ros::master::getPort());
        return genom_mwerr(&d, h);
      }
      ros::start();
    }
    catch (std::exception &e) {
      genom_mwerr_detail d;
      snprintf(d.what, sizeof(d.what), "%s", e.what());
      return genom_mwerr(&d, h);
    }

    if (pipe(ros_evpipe) ||
        fcntl(ros_evpipe[0], F_SETFL, O_NONBLOCK))
      goto error;
  }

  /* create event task */
  if (!ros_evtid_init) {
    s = pthread_create(&ros_evtid, NULL, ros_evtask, NULL);
    if (s) {
      if (!ros_nodeinit) ros::shutdown();
      goto error;
    }
    ros_evtid_init = 1;
  }

  ros_nodeinit++;
  return 0;

error:
  return make_genom_syserr(errno, h);
}


/* --- ros_fini ------------------------------------------------------------ */

void
ros_fini(void)
{
  int i;

  if (ros_nodeinit) ros_nodeinit--;
  if (ros_nodeinit) return;

  ros::shutdown();

  if (ros_evtid_init) {
    pthread_join(ros_evtid, NULL);
    ros_evtid_init = 0;
  }

  for(i=0; i<2; i++) {
    if (ros_evpipe[i] > 0) {
      close(ros_evpipe[i]);
      ros_evpipe[i] = -1;
    }
  }
}


/* --- ros_getopt ---------------------------------------------------------- */

/* options */
static const char shortopts_string[] = "+i:t:";
static struct option longopts_list[] = {
  { "name",		required_argument,	NULL,	'i'},
  { "topic_timeout",	required_argument,	NULL,	't'},
  { NULL, 0, NULL, 0}
};

const struct genom_client_option genom_ros_options[] = {
  { "-i|--name",		"instance name" },
  { "-t|--topic_timeout",
    "timeout subsciptions after that many seconds" },
  { NULL, NULL}
};

genom_event
ros_getopt(genom_client h, int argc, char *argv[], struct ros_options *opts)
{
  int c;

  opts->instance = NULL;
  opts->topics_timeout = 10; /* 10s by default */

  /* reset getopt */
#ifdef HAVE_OPTRESET
  optreset = 1;
#endif
#ifdef __GLIBC__
  optind = 0; /* not "1", see gnu manual for getopt(3) ... */
#else
  optind = 1;
#endif

  /* parse command line options */
  opterr = 0;
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'i': opts->instance = optarg; break;
      case 't': opts->topics_timeout = atoi(optarg); break;

      case '?':
      default:
        genom_mwerr_detail d;
        snprintf(d.what, sizeof(d.what), "unknown option %s", argv[optind]);
        return genom_mwerr(&d, h);
    }

  return 0;
}


/* --- ros_newrqst --------------------------------------------------------- */

genom_event
ros_newrqst(genom_client h, size_t outsize, int &rqstid)
{
  int i, n;
  assert(h);

  rqstid = -1;
  pthread_mutex_lock(&h->activelk);
  n = h->active_rqst.size();
  for(i = h->next; i < h->next + n; i++)
    if (h->active_rqst[i%n].status == ROS_CLIENT_FREE) {
      rqstid = i%n;
      break;
    }
  if (rqstid < 0) {
    try {
      h->active_rqst.resize(n+1);
    } catch(...) { return make_genom_syserr(ENOMEM, h); }

    rqstid = n;
  }
  h->next = rqstid + 1;

  h->active_rqst[rqstid].info = NULL;
  h->active_rqst[rqstid].status = ROS_CLIENT_WAIT_ACK;
  h->active_rqst[rqstid].pstatus = ROS_CLIENT_WAIT_ACK;
  h->active_rqst[rqstid].error = genom_ok;
  h->active_rqst[rqstid].exdetail = NULL;
  h->active_rqst[rqstid].exfini = NULL;
  h->active_rqst[rqstid].output = malloc(outsize);
  h->active_rqst[rqstid].sentcb = NULL;
  h->active_rqst[rqstid].donecb = NULL;
  h->active_rqst[rqstid].cb_data = NULL;
  if (!h->active_rqst[rqstid].output) {
    h->active_rqst[rqstid].status = ROS_CLIENT_FREE;
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOMEM, h);
  }

  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_eventfd --------------------------------------------------------- */

int
ros_eventfd(void)
{
  return ros_evpipe[0];
}


/* --- ros_service_info ---------------------------------------------------- */

const struct genom_service_info *
ros_service_info(genom_client h, int rqstid)
{
  const struct genom_service_info *info = NULL;

  pthread_mutex_lock(&h->activelk);
  if (rqstid >= 0 && rqstid < (signed)h->active_rqst.size())
    if (h->active_rqst[rqstid].status != ROS_CLIENT_FREE)
      info = h->active_rqst[rqstid].info;
  pthread_mutex_unlock(&h->activelk);

  return info;
}


/* --- ros_waitevent ------------------------------------------------------- */

genom_event
ros_waitevent(genom_client h)
{
  fd_set fdset;
  int s;

  FD_ZERO(&fdset);
  do {
    FD_SET(ros_evpipe[0], &fdset);
    s = select(ros_evpipe[0]+1, &fdset, NULL, NULL, NULL);
  } while(s < 0 && errno == EINTR);

  if (s < 0) return make_genom_syserr(errno, h);
  return genom_ok;
}


/* --- ros_waitack --------------------------------------------------------- */

genom_event
ros_waitack(genom_client h, int rqstid)
{
  genom_event s;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= (signed)h->active_rqst.size()) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT, h);
  }
  while(h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_ACK) {
    pthread_mutex_unlock(&h->activelk);

    s = ros_waitevent(h);
    if (!s) s = ros_doevents(h);
    if (s) return s;

    pthread_mutex_lock(&h->activelk);
  }

  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_waitterm -------------------------------------------------------- */

genom_event
ros_waitterm(genom_client h, int rqstid)
{
  genom_event s;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= (signed)h->active_rqst.size()) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT, h);
  }

  while(h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_ACK ||
        h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_TERM) {
    pthread_mutex_unlock(&h->activelk);

    s = ros_waitevent(h);
    if (!s) s = ros_doevents(h);
    if (s) return s;

    pthread_mutex_lock(&h->activelk);
  }

  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_done ------------------------------------------------------------ */

int
ros_done(genom_client h, int rqstid)
{
  int done = 1;

  pthread_mutex_lock(&h->activelk);
  if (rqstid >= 0 || rqstid < (signed)h->active_rqst.size())
    if (h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_ACK ||
        h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_TERM)
      done = 0;
  pthread_mutex_unlock(&h->activelk);

  return done;
}


/* --- ros_result ---------------------------------------------------------- */

genom_event
ros_result(genom_client h, int rqstid,
           genom_event *report, void **output, void **exdetail)
{
  genom_client_s::request *a;
  genom_event s;

  s = ros_waitterm(h, rqstid);
  if (s) return s;

  pthread_mutex_lock(&h->activelk);
  a = &h->active_rqst[rqstid];
  if (a->status == ROS_CLIENT_FREE) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT, h);
  }

  if (report) *report = a->error;
  if (output) *output = a->output;
  if (exdetail) *exdetail = a->exdetail;
  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_clean ----------------------------------------------------------- */

genom_event
ros_clean(genom_client h, int rqstid)
{
  genom_client_s::request *a;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= (signed)h->active_rqst.size() ||
      h->active_rqst[rqstid].status == ROS_CLIENT_FREE) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT, h);
  }
  a = &h->active_rqst[rqstid];

  if (a->status == ROS_CLIENT_WAIT_ACK || a->status == ROS_CLIENT_WAIT_TERM) {
    genom_event s;
    genom_gh gh = a->gh; /* gh cannot be used with the activelk held */
    pthread_mutex_unlock(&h->activelk);
    gh.cancel();
    s = ros_waitterm(h, rqstid);
    if (s) return s;
    return ros_clean(h, rqstid);
  }

  if (a->output) {
    if (a->info->fini_output) a->info->fini_output(a->output);
    free(a->output);
  }
  if (a->exdetail) {
    if (a->exfini) a->exfini(a->exdetail);
    free(a->exdetail);
  }
  a->status = ROS_CLIENT_FREE;
  pthread_mutex_unlock(&h->activelk);

  return genom_ok;
}


/* --- ros_doevents -------------------------------------------------------- */

genom_event
ros_doevents(genom_client h)
{
  struct genom_client_s::request *a;
  ssize_t r;
  char e[16];
  unsigned int i;

  if (ros_evpipe[0] < 0)
    return make_genom_syserr(EBADF, h);

  do {
    do {
      r = read(ros_evpipe[0], e, sizeof(e));
    } while (r == -1 && errno == EINTR);
    if (r == 0) return make_genom_syserr(EBADF, h);

    if (r == -1) {
      if ((errno == EAGAIN || errno == EWOULDBLOCK)) break;
      return make_genom_syserr(errno, h);
    }

    for(i=0; (signed)i<r; i++) {
      switch(e[i]) {
        case ros_end:
          return make_genom_syserr(ENODEV, h);

        case ros_cbevent:
          break;

        default: assert(!"reached unreachable");
      }
    }
  } while(r == sizeof(e));

  pthread_mutex_lock(&h->activelk);
  if (!h->events) {
    pthread_mutex_unlock(&h->activelk);
    return genom_ok;
  }
  h->events = 0;
  for(i = 0; i < h->active_rqst.size(); i++) {
    a = &h->active_rqst[i];
    if (a->status == a->pstatus) continue;

    switch(a->status) {
      case ROS_CLIENT_WAIT_TERM:
        a->pstatus = a->status;
        if (a->sentcb)
          (*a->sentcb)(h, i, 0, a->error, NULL, a->exdetail, a->cb_data);
        break;

      case ROS_CLIENT_TERM:
        a->pstatus = a->status;
        if (a->donecb) {
          if (a->error)
            (*a->donecb)(h, i, 1, a->error, NULL, a->exdetail, a->cb_data);
          else
            (*a->donecb)(h, i, 1, genom_ok, a->output, NULL, a->cb_data);
        }
        break;
    }
  }
  pthread_mutex_unlock(&h->activelk);

  return genom_ok;
}


/* --- exceptions ---------------------------------------------------------- */

genom_event
ros_client_raise(genom_event ex, void *detail, size_t size, genom_context self)
{
  genom_client h = self->data->h;

  if (h->exsize < size) {
    void *exdetail = realloc(h->exdetail, size);
    if (!exdetail) return h->ex = genom_fatal;

    h->exsize = size;
    h->exdetail = exdetail;
  }
  h->ex = ex;
  memcpy(h->exdetail, detail, size);
  return ex;
}

const void *
ros_client_raised(genom_event *ex, genom_context self)
{
  genom_client h = self->data->h;

  if (ex) *ex = h->ex;
  return h->exdetail;
}


/* --- fake callback for services ------------------------------------------ */

static struct ros_fake_cb_ : public ros::CallbackInterface {
  CallResult call(void) { return Success; }
} ros_fake_cb;

struct ros_fake_delete {
  void operator()(ros_fake_cb_ *) {}
};

void
ros_evtask_signal(void)
{
  ros::CallbackInterfacePtr cb(&ros_fake_cb, ros_fake_delete());
  ros::getGlobalCallbackQueue()->addCallback(cb);
}


/* --- ros_evtask ---------------------------------------------------------- */

static void *
ros_evtask(void *data)
{
  ros::CallbackQueue *queue = ros::getGlobalCallbackQueue();
  char c;

  c = ros_cbevent;
  while(ros::ok()) {
    queue->callAvailable(ros::WallDuration(10.));
    if (!ros_nodeinit) break;
    write(ros_evpipe[1], &c, 1);
  }

  c = ros_end;
  write(ros_evpipe[1], &c, 1);
  return NULL;
}
